import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;

import opennlp.tools.sentdetect.SentenceDetectorME;
import opennlp.tools.sentdetect.SentenceModel;
import opennlp.tools.util.Span;


/**
 * Classe responsavel pela separacao (Split) das sentencas de um determinado texto.
 * Utiliza-se da biblioteca OpenNLP (http://incubator.apache.org/opennlp/) que por sua vez
 * utiliza-se de processamento estatistico (Maxima Entropia).
 */
public class SentDetector {

	private String caminhoArquivoBIN;
	

	/**
	 * Contrutor da classe.
	 * @param caminhoArquivoBIN Endereco do arquivo .bin (pt-sent.bin, por exemplo).
	 */
	public SentDetector(String caminhoArquivoBIN){
		this.caminhoArquivoBIN = caminhoArquivoBIN;
		System.out.println("Construindo Detector de Sentencas...");
	}
	

	public String getCaminhoArquivoBIN() {
		return caminhoArquivoBIN;
	}

	public void setCaminhoArquivoBIN(String caminhoArquivoBIN) {
		this.caminhoArquivoBIN = caminhoArquivoBIN;
	}
	
	
	/**
	 * Divide o texto passado como parametro em suas respectivas 
	 * sentencas.
	 * 
	 * @param modelo Modelo a ser utilizado (dependente da linguagem)
	 * @param texto Texto a ser dividido. 
	 * @return Lista contendo as frases componentes do texto.
	 */
	public String[] getSentencas(String texto){
		SentenceModel modelo = this.getModelo(this.getCaminhoArquivoBIN());
		SentenceDetectorME sd = new SentenceDetectorME(modelo);
		String[] sentencas = sd.sentDetect(texto);
                System.out.println("TESTE: "+sd.sentDetect(texto).length);
		return(sentencas);
	}
	
	/**
	 * Retorna os limites (inico...fim) das sentencas obtidas mediante quebra do texto.
	 * @param modelo Modelo (dependente de linguagem) utilizado (pt-sent.bin, por exemplo);
	 * @param texto Texto a ser dividido.
	 * @return Lista contendo os limites.
	 */
	public Span[] getLimSent(String texto){
		SentenceModel modelo = this.getModelo(this.getCaminhoArquivoBIN());
		SentenceDetectorME sd = new SentenceDetectorME(modelo);
		Span[] tamanhos = sd.sentPosDetect(texto);
		return(tamanhos);
	}
	
	/**
	 * Retorna a probabilidade das sentencas obtidas mediante quebra do texto.
	 * @param modelo Modelo (dependente de linguagem) utilizado (pt-sent.bin, por exemplo);
	 * @param texto Texto a ser dividido.
	 * @return Lista contendo as probabilidades.
	 */
	public double[] getProbSent(String texto){
		SentenceModel modelo = this.getModelo(this.getCaminhoArquivoBIN());
		SentenceDetectorME sd = new SentenceDetectorME(modelo);
		sd.sentDetect(texto);
		double[] prob = sd.getSentenceProbabilities();
		return(prob);
	}
	
	/**
	 * Retorna o modelo de sentença.
	 * @param caminhoArquivoBIN Caminho ate o arquivo .bin (pt-sent.bin, por exemplo)
	 * @return Modelo de Sentença.
	 */
	public SentenceModel getModelo(String caminhoArquivoBIN){
		return (Util.getModeloSent(caminhoArquivoBIN));
	}
	
}
